@@ -14,6 +14,8 @@ from web3.contract import ContractEvent
 class Filter:
     """Internal filter used to match events.
 
+    A helper class to deal with `eth_getLogs` JSON API.
+
     This can be used with
 
     - Historical events: :py:mod:`eth_defi.event_reader.reader`
@@ -28,11 +30,13 @@ class Filter:
     #: TODO: Currently unsupported
     bloom: Optional[BloomFilter]
 
-    #: Get events from a single contract only
-    contract_address: Optional[str] = None
+    #: Get events from a single contract only.
+    #:
+    #: For multiple contracts give a list of addresses.
+    contract_address: Optional[str | List[str]] = None
 
     @staticmethod
-    def create_filter(address: Optional[str], event_types: List[Type[ContractEvent]]) -> "Filter":
+    def create_filter(address: Optional[str | List[str]], event_types: List[Type[ContractEvent]]) -> "Filter":
 
         topics = {event_type.build_filter().topics[0]: event_type for event_type in event_types}
 
